(* On utilise Camlp4, donc:
 
   i) en mode interprete, on utilise 
 
       ocaml camlp4o.cma pa_extend.cmo q_MLast.cmo extend_monade.cmo

  ii) en mode compile, on utilise

       ocamlc -pp "camlp4o pa_extend.cmo q_MLast.cmo extend_monade.cmo" -I +camlp4 -c interp-monad-evolution.ml

 iii) comme preprocessing, on ecrit

       camlp4o pa_extend.cmo q_MLast.cmo ./extend_monade.cmo pr_o.cmo interp-monad-evolution.ml

 *)

(*                                                                      *)
(* Semantique Denotationnelle comme Interpretation de TFun              *)
(*                                                                      *)
(* Roberto Di Cosmo (roberto@dicosmo.org)            2004               *)
(*                                                                      *)
(* Syntaxe Abstraite du Langage                                         *)
(*                                                                      *)

type tfuntype  = TBool | TNat | Arr of tfuntype * tfuntype   (* les types       *)

type  id        = string                                      (* identificateurs *) 

type  op         = Plus
              | Minus
	      | Mult
	      | Div

type  compop     = Eq
              | Less
	      | More

type  exp       = INT of int | BOOL of bool                   (* les expressions *)
              | Iden of id
              | App of (exp * exp)
              | Abs of ((id * tfuntype) * exp)
              | IF of (exp * exp * exp)
              | Compop  of compop * exp * exp
              | Base of (op * exp * exp)
              | Rec of (id * id * exp)                        (* recursion      *)
              | Raise                                         (* error          *)
              | Lkp of exp                                    (* effets de bord *)
              | Ref of exp
              | Upd of exp * exp
              | Seq of exp * exp


(* les operation de base *)

let getop = function
    Plus -> fun x y -> x+y
  | Minus -> fun x y -> x-y
  | Mult -> fun x y -> x*y
  | Div -> fun x y -> x/y
;;

let getcompop = function
    Eq -> fun x y -> x=y
  | Less -> fun x y -> x<y
  | More -> fun x y -> x>y
;;


(*                                                                           *)
(*            TFun : l'Interpreteur                                          *)
(*                                                                           *)
            
(* Ce code est une traduction en ML immediate de la Semantique               *)
(* Denotationnelle Directe de TFun dans FTH qu'on a donne' en cours          *)
(* Un grande avantage de cet approche est la clarte du code,                 *)
(* qui se commente tout seul                                                 *)
(* On remarquera cependant que on n'utilise pas ici les types du langage au  *)
(* contraire de ce que l'on fait dans la semantique mathematique.            *)
(* En effet, pour simplicite, on n'utilise pas ici la derivation de typage   *)
(* pour la definition inductive, mais seulement la structure des programmes. *)
(* Cela ne change pas vraiement les fonctions semantiques, mais  on          *)
(* accepte comme parametre des programmes qui sont mal types, ce qui         *)
(* apparait en trois points: l'erreur possible dans projectvalue et les deux *)
(* match incomplets pour Arrow et Bool. En effet, sans derivation de typage, *)
(* on ne peut pas garantir que le programme passe en parametre fournisse les *)
(* bons valeurs dans ces cas                                                 *)            

(* monads and monad transformers *)

(* for each relevant computational type, we give here the associated monad and monad transformers *)

module type Monad = 
   sig
     type 'a t
     val tval: 'a -> 'a t
     val tlift: ('a -> 'b t) -> ('a t -> 'b t)
         (* to unpack the computation *)
     val reveal: 'a t -> 'a

     val raise: unit -> 'a t  (* for errors *)

     val lkp  : int -> int t    (* for state  *)
     val update: int -> int -> unit t
     val init : int -> int t 
   end

module Id =
  struct
    type 'a t = 'a
    let tval v = v
    let tlift f = f
    let reveal v = v

    (* undefined operations *)

    let raise ()  = failwith "Id: no raise operation in Id monad"

    let update l  = failwith "Id: undefined operation update"
    let lkp l     = failwith "Id: undefined operation lkp"
    let init v    = failwith "Id: undefined operation init"
  end

module Lifting = 
  struct
    type 'a t = Up of 'a | Bot
    let tval v = Up v
    let tlift f e = match e with Bot -> Bot | Up v -> f v
    let reveal lv = match lv with Bot -> failwith "Lifting: cannot project Bot" | Up v -> v

    (* undefined operations *)

    let raise ()  = failwith "Lifting: no raise operation in Lifting monad"

    let update l  = failwith "Lifting: undefined operation update"
    let lkp l     = failwith "Lifting: undefined operation lkp"
    let init v    = failwith "Lifting: undefined operation init"
  end

module ILifting (T: Monad) = (* this is not really a monad transformer, as the T in the parameter is used inside 'a t.
                                but I leave it here for future reference, as I need to understand why the order
                                of usage of T is relevant *)
  struct
    type 'a t = Up of 'a T.t | Bot
    let tval v = Up (T.tval v)
    let tlift f e = match e with Bot -> Bot | Up v -> Up (T.tlift f v)
    let reveal lv = match lv with Bot -> failwith "Lifting: cannot project Bot" | Up v -> T.reveal v

    (* undefined operations *)

    let raise ()  = failwith "Lifting: no raise operation in Lifting monad"

    let update l  = failwith "Lifting: undefined operation update"
    let lkp l     = failwith "Lifting: undefined operation lkp"
    let init v    = failwith "Lifting: undefined operation init"
  end

module Exc =
  struct
    type 'a t = Res of 'a | Err
    let tval v = Res v
    let tlift f e = match e with Err -> Err | Res v -> f v
    let reveal lv = match lv with Err -> failwith "Error: cannot project Err" | Res v -> v

    let raise () = Err

    (* for state *)

    let update l  = failwith "Exc: undefined operation update"
    let lkp l     = failwith "Exc: undefined operation lkp"
    let init v    = failwith "Exc: undefined operation init"

  end

module IExc (T: Monad) =
  struct
    type 'a maybe = Res of 'a | Err
    type 'a t = 'a maybe T.t
    let tval v = T.tval (Res v)
    let tlift f e = T.tlift (fun m -> match m with Res v -> f v | Err -> T.tval Err) e
    let reveal lv = match (T.reveal lv) with Err -> failwith "Error: cannot project Err" | Res v -> v

    (* raising errors *)

    let raise () = T.tval Err

    (* redefinition of memory manipulations: since only the type changes, we need not really redefine things deeply *)
    let tmap f e = T.tlift (fun x -> T.tval (f x)) e  (* operation derivee *)

    let update l v = tmap (fun x -> Res x) (T.update l v)
    let lkp l = tmap (fun x -> Res x) (T.lkp l)
    let init v = tmap (fun x -> Res x) (T.init v)

  end

module State = 
  struct
    type loc = int
    type sv = int
    type s = loc -> sv 
    type 'a t = s -> 'a * s
    let tval (v: 'a) = (fun s -> (v,s) : 'a t)
    let tlift (f: 'a -> 'b t) (e: 'a t) = 
        ((fun s -> let (v,s') = e s 
                   in f v s') : 'b t)

    (* Uninitialised memory *)

    let emptys = (fun l -> failwith "State: access to uninitialised memory location")

    let reveal texp = let (v,s') = texp emptys in v

    (* raise *)

    let raise () = failwith "State: no raise operation in State monad"

    (* newloc, update, lkp, init *)

    let newloc = let l = ref 0 in fun () -> let nl = !l in l:= nl+1; nl
    let update (l:loc) (v:sv) s = let s' = fun l' -> if l = l' then v else s l' in ((),s')
    let lkp (l:loc) = fun s -> ((s l:sv),s)

    let init (v:sv) = fun s -> let (l:loc) = newloc() in 
                          let s' = fun l' -> if l = l' then v else s l' in (l, s')
  end

module IState (T: Monad) =
  struct
    type loc = int
    type sv = int
    type s = loc -> sv
    type 'a t = s -> ('a * s) T.t 
    let tval v = fun s -> T.tval (v,s)
    let tlift f e = fun s -> T.tlift (fun (v',s') -> f v' s') (e s)

    (* Uninitialised memory *)

    let emptys = (fun l -> failwith "State: access to uninitialised memory location")

    let reveal texp = let (v,s') = T.reveal (texp emptys) in v

    (* auxiliary function newloc *)

    let newloc = let l = ref 0 in fun () -> let nl = !l in l:= nl+1;nl

    (* update, lkp, init *)

    let update l (v:sv) (s:s) = let (s':s) = fun l' -> if l = l' then v else s l' in T.tval((),s')
    let lkp l = fun (s:s) -> T.tval((s l:sv),s)
    let init (v:sv) = fun (s:s) -> let l = newloc() in 
                          let (s':s) = fun l' -> if l = l' then v else s l' in
                          T.tval(l, s')

    (* redefinition of raise *)

    let raise () = fun s -> T.raise()

  end
    
(**** La construction de l'interprete *)

module Interp (T:Monad) =
struct

(* Definition des Domaines Semantiques:                                      *)

type cpo = (* les valeurs de base: ici les entiers et booleens *)
           Int of int | Bool of bool 
           (* les constructions du domaine *)           
	 | Unit of unit
         | Arrow of (idt -> expt)
and  idt  = cpo                (* CBV *) 
and  expt = cpo T.t

(* ouverture de la definition de la monade *)

let (tval: idt -> expt) = T.tval

let tlift (f:idt -> expt) (texp: expt) = T.tlift f texp

let tmap f e = T.tlift (fun x -> T.tval (f x)) e  (* operation derivee *)

(* extension pour les locations *)

let injint n = Int n
let injunit () = Unit ()

let lkp_T l = tmap injint (T.lkp l)
let init_T v  = tmap injint (T.init v)
let upd_T l v = tmap injunit (T.update l v)

(* ceci est un point fixe pour le CBV de ML *)

let rec fix f = f (fun x -> (fix f) x)

(* acceder a l'environnement *)

let rec proj (i,(env: (id * idt) list)) =
      match env with
        []        -> failwith "Identificateur non trouve"
      | (id,a)::r -> if id = i then a else proj (i, r)

(* la fonction semantique *)

let rec (esem: exp -> (id * idt) list -> expt) = fun  exp env ->
        match exp with
          App (exp1,exp2) -> let_T (Arrow f) <= esem exp1 env in                                         (* CBV *)
	                     let_T v <= esem exp2 env in
                              f v
        | Abs ((id,t),exp)-> tval(Arrow(fun a -> (esem exp ((id,a)::env))))
        | Iden(id)        -> tval(proj(id,env))
        | INT i           -> tval(Int i)
        | BOOL b          -> tval(Bool b)
        | IF(b,exp1,exp2) -> let_T Bool bv <= esem b env in
                             if bv then esem exp1 env
                             	   else esem exp2 env
        | Compop(op,exp1,exp2) -> let_T Int n <= esem exp1 env in
                                  let_T Int m <= esem exp2 env in
                                    if (getcompop op n m) then tval(Bool true)
                                                          else tval(Bool false)
        | Base(op,exp1,exp2) -> let_T Int n <= esem exp1 env in
                                let_T Int m <= esem exp2 env in
                                 tval (Int ((getop op) n m))
        | Rec(f,x,exp)    -> tval(Arrow(fix(fun g -> (fun a -> (esem exp ((x,a)::(f,Arrow g)::env))))))  (* CBV *)
        | Raise           -> T.raise()
        | Lkp e           -> let_T (Int l) <= esem e env in lkp_T l
        | Ref e           -> let_T (Int v) <= esem e env in init_T v
        | Upd (e1,e2)     -> let_T (Int l) <= esem e1 env in
                             let_T (Int v) <= esem e2 env in
                                upd_T l v

end


(* Maintenant, on teste! *)

module IntId = Interp Id

module IntS = Interp State

module IntE = Interp Exc


module ES = IExc State

module IntES = Interp ES

module SE = IState Exc

module IntSE = Interp SE


(* ici, on peut executer ces programmes apres avoir fait open du bon interprete *)


(* Programme bien type *)
let r1 = esem (App(Abs(("x",TBool),IF(Iden "x",Base(Plus,INT 1,INT 5),INT 2)),BOOL true)) [];;

(* Programme mal type *)
let r2 = esem (App(INT 1, INT 2))[];;

(* Factoriel *)

let r3 = esem (App(Rec("fact","x",IF(Compop(Eq,Iden "x",INT 0),INT 1,Base(Mult,Iden "x", App(Iden "fact",Base(Minus,Iden "x",INT 1))))),INT 10)) [];;

(* Memoire *)

let r4 = (esem (App(Abs(("x",TBool),Lkp (Iden "x")),Ref(INT 33))) []);;



